<?php
namespace nur\sery\db\sqlite;

use nur\sery\cl;
use nur\sery\php\func;
use nur\sery\ValueException;

/**
 * Class SqliteCapacitor: objet permettant d'accumuler des données pour les
 * réutiliser plus tard
 */
class SqliteCapacitor {
  function __construct($sqlite) {
    $this->sqlite = Sqlite::with($sqlite);
  }

  /** @var Sqlite */
  protected $sqlite;

  function sqlite(): Sqlite {
    return $this->sqlite;
  }

  protected function getTableName(?string $channel): string {
    return ($channel ?? "default")."_channel";
  }

  /** tester si le canal spécifié existe */
  function exists(?string $channel=null): bool {
    #XXX maintenir une table channels avec la liste des canaux valides
  }

  /** supprimer le canal spécifié */
  function reset(?string $channel=null) {
    $tableName = $this->getTableName($channel);
    $this->sqlite->exec("drop table if exists $tableName");
    #XXX maj de la tables channels dans une transaction
  }

  /** @var array */
  protected $created;

  protected function getKeyDefinitions(?string $channel): ?array {
    return null;
  }

  protected function getKeyValues($item, ?string $channel): ?array {
    return null;
  }

  protected function create(?string $channel): void {
    if ($this->created[$channel] ?? false) return;
    $columns = cl::merge([
      "_id" => "integer primary key autoincrement",
      "_item" => "text",
    ], $this->getKeyDefinitions($channel));
    #XXXvvv migrer vers la syntaxe tableau de create
    $index = 0;
    foreach ($columns as $column => &$definition) {
      if ($column === $index) {
        $index++;
      } else {
        $definition = "$column $definition";
      }
    }; unset($definition);
    $query = implode("", [
      "create table if not exists ",
      $this->getTableName($channel),
      " (",
      implode(", ", $columns),
      ")",
    ]);
    #XXX^^^ migrer vers la syntaxe tableau de create
    $this->sqlite->exec($query);
    #XXX maj de la tables channels dans une transaction
    $this->created[$channel] = true;
  }

  function charge($item, ?string $channel=null) {
    $this->create($channel);
    $values = cl::merge($this->getKeyValues($item, $channel), [
      "_item" => serialize($item),
    ]);
    $this->sqlite->exec([
      "insert",
      "into" => $this->getTableName($channel),
      "values" => $values,
    ]);
  }

  /** décharger les données du canal spécifié */
  function discharge($keys=null, ?string $channel=null, ?bool $reset=null): iterable {
    if ($keys !== null && !is_array($keys)) $keys = ["_id" => $keys];
    if ($reset === null) $reset = $keys === null;
    $rows = $this->sqlite->all([
      "select _item",
      "from" => $this->getTableName($channel),
      "where" => $keys,
    ]);
    foreach ($rows as $row) {
      $item = unserialize($row['_item']);
      yield $item;
    }
    if ($reset) $this->reset($channel);
  }

  /**
   * obtenir l'élément identifié par les clés spécifiées sur le canal spécifié
   *
   * si $keys n'est pas un tableau, il est transformé en ["_id" => $keys]
   */
  function get($keys, ?string $channel=null) {
    if ($keys === null) throw ValueException::null("keys");
    if (!is_array($keys)) $keys = ["_id" => $keys];
    $row = $this->sqlite->one([
      "select _item",
      "from" => $this->getTableName($channel),
      "where" => $keys,
    ]);
    if ($row === null) return null;
    else return unserialize($row["_item"]);
  }

  /**
   * appeler une fonction pour chaque élément du canal spécifié.
   *
   * $keys permet de filtrer parmi les élements chargés
   *
   * si $func retourne un tableau, il est utilisé pour mettre à jour
   * l'enregistrement.
   */
  function each($keys, callable $func, ?array $args=null, ?string $channel=null): void {
    if ($keys !== null && !is_array($keys)) $keys = ["_id" => $keys];
    $context = func::_prepare($func);
    $sqlite = $this->sqlite;
    $tableName = $this->getTableName($channel);
    $rows = $sqlite->all([
      "select",
      "from" => $tableName,
      "where" => $keys,
    ]);
    $args ??= [];
    foreach ($rows as $row) {
      $item = unserialize($row['_item']);
      $updates = func::_call($context, [$item, $row, ...$args]);
      if (is_array($updates)) {
        if (array_key_exists("_item", $updates)) {
          $updates["_item"] = serialize($updates["_item"]);
        }
        #XXXvvv migrer vers la syntaxe tableau de update
        $params = null;
        $sql = ["update", $tableName, "set"];
        _query::parse_set_values($updates, $setsql, $params);
        $sql[] = implode(", ", $setsql);
        $sql[] = "where";
        _query::parse_conds(["_id" => $row["_id"]], $wheresql, $params);
        $sql[] = implode(" and ", $wheresql);
        $sqlite->exec(implode(" ", $sql), $params);
        #XXX^^^ migrer vers la syntaxe tableau de update
      }
    }
  }

  function close(): void {
    $this->sqlite->close();
  }
}
